from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
import math

import nngen.util as util


def conv2d(input, filter, strides,
           bias=None, scale=None,
           rshift_mul=None, rshift_sum=None, rshift_out=None,
           act_func=None, padding='SAME',
           asymmetric_clip=False,
           dtype=None, mul_dtype=None, sum_dtype=None,
           name=None,
           par_ich=1, par_och=1, par_col=1, par_row=1,
           concur_och=None, stationary='filter',
           input_ram_size=None, filter_ram_size=None,
           bias_ram_size=None, scale_ram_size=None,
           vshamt_mul_ram_size=None, vshamt_sum_ram_size=None, vshamt_out_ram_size=None,
           out_ram_size=None,
           disable_keep_input=False,
           input_shape=None, filter_shape=None, out_shape=None,
           input_dtype=None, filter_dtype=None,
           bias_dtype=None, scale_dtype=None,
           vshamt_mul_dtype=None, vshamt_sum_dtype=None, vshamt_out_dtype=None):

    # opposite order to pool
    if isinstance(padding, str) and padding == 'SAME':
        pad_col, pad_col_right, pad_col_left = util.pad_size_split(
            input.shape[2], filter.shape[2], strides[2])
        pad_row, pad_row_bottom, pad_row_top = util.pad_size_split(
            input.shape[1], filter.shape[1], strides[1])

        shape = (int(math.ceil(input.shape[0] / strides[0])),
                 util.pix_size(input.shape[1],
                               filter.shape[1], strides[1], 'SAME'),
                 util.pix_size(input.shape[2],
                               filter.shape[2], strides[2], 'SAME'),
                 int(math.ceil(filter.shape[0] / strides[3])))

    elif isinstance(padding, str) and padding == 'VALID':
        pad_col, pad_col_right, pad_col_left = 0, 0, 0
        pad_row, pad_row_bottom, pad_row_top = 0, 0, 0

        shape = (int(math.ceil(input.shape[0] / strides[0])),
                 util.pix_size(input.shape[1],
                               filter.shape[1], strides[1], 'VALID'),
                 util.pix_size(input.shape[2],
                               filter.shape[2], strides[2], 'VALID'),
                 int(math.ceil(filter.shape[0] / strides[3])))

    elif isinstance(padding, int):
        pad_col, pad_col_right, pad_col_left = padding * 2, padding, padding
        pad_row, pad_row_bottom, pad_row_top = padding * 2, padding, padding

        shape = (int(math.ceil(input.shape[0] / strides[0])),
                 util.pix_size(input.shape[1] + padding * 2,
                               filter.shape[1], strides[1], 'VALID'),
                 util.pix_size(input.shape[2] + padding * 2,
                               filter.shape[2], strides[2], 'VALID'),
                 int(math.ceil(filter.shape[0] / strides[3])))

    elif isinstance(padding, (tuple, list)):
        pad_col, pad_col_right, pad_col_left = padding[2] + padding[3], padding[3], padding[2]
        pad_row, pad_row_bottom, pad_row_top = padding[0] + padding[1], padding[1], padding[0]

        shape = (int(math.ceil(input.shape[0] / strides[0])),
                 util.pix_size(input.shape[1] + padding[0] + padding[1],
                               filter.shape[1], strides[1], 'VALID'),
                 util.pix_size(input.shape[2] + padding[2] + padding[3],
                               filter.shape[2], strides[2], 'VALID'),
                 int(math.ceil(filter.shape[0] / strides[3])))
    else:
        raise ValueError("padding options must be 'SAME', 'VALID', int, tuple, or list.")

    out = np.zeros(shape, dtype=np.int64)

    input = np.pad(input, [(0, 0),
                           (pad_row_top, pad_row_bottom),
                           (pad_col_left, pad_col_right),
                           (0, 0)], 'constant')

    if bias is None:
        bias = np.zeros([shape[-1]], dtype=np.int64)
    elif not isinstance(bias, np.ndarray):
        new_bias = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_bias.shape[-1]):
            new_bias[i] = bias
        bias = new_bias
    elif len(bias.shape) == 1 and bias.shape[0] == 1:
        new_bias = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_bias.shape[-1]):
            new_bias[i] = bias[0]
        bias = new_bias

    if scale is None:
        scale = np.ones([shape[-1]], dtype=np.int64)
    elif not isinstance(scale, np.ndarray):
        new_scale = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_scale.shape[-1]):
            new_scale[i] = scale
        scale = new_scale
    elif len(scale.shape) == 1 and scale.shape[0] == 1:
        new_scale = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_scale.shape[-1]):
            new_scale[i] = scale[0]
        scale = new_scale

    if rshift_mul is None:
        rshift_mul = np.zeros([shape[-1]], dtype=np.int64)
    elif not isinstance(rshift_mul, np.ndarray):
        new_rshift_mul = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_rshift_mul.shape[-1]):
            new_rshift_mul[i] = rshift_mul
        rshift_mul = new_rshift_mul
    elif len(rshift_mul.shape) == 1 and rshift_mul.shape[0] == 1:
        new_rshift_mul = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_rshift_mul.shape[-1]):
            new_rshift_mul[i] = rshift_mul[0]
        rshift_mul = new_rshift_mul

    if rshift_sum is None:
        rshift_sum = np.zeros([shape[-1]], dtype=np.int64)
    elif not isinstance(rshift_sum, np.ndarray):
        new_rshift_sum = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_rshift_sum.shape[-1]):
            new_rshift_sum[i] = rshift_sum
        rshift_sum = new_rshift_sum
    elif len(rshift_sum.shape) == 1 and rshift_sum.shape[0] == 1:
        new_rshift_sum = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_rshift_sum.shape[-1]):
            new_rshift_sum[i] = rshift_sum[0]
        rshift_sum = new_rshift_sum

    if rshift_out is None:
        rshift_out = np.zeros([shape[-1]], dtype=np.int64)
    elif not isinstance(rshift_out, np.ndarray):
        new_rshift_out = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_rshift_out.shape[-1]):
            new_rshift_out[i] = rshift_out
        rshift_out = new_rshift_out
    elif len(rshift_out.shape) == 1 and rshift_out.shape[0] == 1:
        new_rshift_out = np.zeros([shape[-1]], dtype=np.int64)
        for i in range(new_rshift_out.shape[-1]):
            new_rshift_out[i] = rshift_out[0]
        rshift_out = new_rshift_out

    rshift_mul_pow = np.where(rshift_mul > np.zeros_like(rshift_mul, dtype=np.int64),
                              rshift_mul - 1,
                              np.zeros_like(rshift_mul))
    rshift_mul_round = np.where(rshift_mul > np.zeros_like(rshift_mul, dtype=np.int64),
                                np.power(np.ones_like(rshift_mul, dtype=np.int64) * 2,
                                         rshift_mul_pow),
                                np.zeros_like(rshift_mul, dtype=np.int64))

    rshift_sum_pow = np.where(rshift_sum > np.zeros_like(rshift_sum, dtype=np.int64),
                              rshift_sum - 1,
                              np.zeros_like(rshift_sum))
    rshift_sum_round = np.where(rshift_sum > np.zeros_like(rshift_sum, dtype=np.int64),
                                np.power(np.ones_like(rshift_sum, dtype=np.int64) * 2,
                                         rshift_sum_pow),
                                np.zeros_like(rshift_sum, dtype=np.int64))

    rshift_out_pow = np.where(rshift_out > np.zeros_like(rshift_out, dtype=np.int64),
                              rshift_out - 1,
                              np.zeros_like(rshift_out))
    rshift_out_round = np.where(rshift_out > np.zeros_like(rshift_out, dtype=np.int64),
                                np.power(np.ones_like(rshift_out, dtype=np.int64) * 2,
                                         rshift_out_pow),
                                np.zeros_like(rshift_out, dtype=np.int64))

    input_point = 0 if input_dtype is None else input_dtype.point
    filter_point = 0 if filter_dtype is None else filter_dtype.point
    bias_point = 0 if bias_dtype is None else bias_dtype.point
    scale_point = 0 if scale_dtype is None else scale_dtype.point
    out_point = (max(input_point, filter_point)
                 if dtype is None else dtype.point)
    out_width = 32 if dtype is None else dtype.width

    mul_point = max(input_point, filter_point)
    mul_shift = min(input_point, filter_point)
    sum_point = mul_point
    add_point = max(sum_point, bias_point)

    sum_shift = add_point - sum_point
    bias_shift = add_point - bias_point
    shifted_bias = np.left_shift(bias, bias_shift)

    scl_point = max(sum_point, scale_point)
    scl_shift = min(sum_point, scl_point)
    shifted_scale = np.right_shift(scale, scl_shift)

    p_th = (1 << (out_width - 1)) - 1
    if asymmetric_clip:
        n_th = -1 * p_th - 1
    else:
        n_th = -1 * p_th

    p_th = p_th >> out_point
    n_th = n_th >> out_point

    def my_matmul_by_matmul(a, w):
        return np.matmul(a, w.T)

    def my_matmul_by_multiply(a, w):
        mul = np.multiply(a, w)
        mul = np.add(mul, rshift_mul_round.reshape([rshift_mul_round.shape[-1], 1]))
        mul = np.right_shift(mul, mul_shift)
        mul = np.right_shift(mul, rshift_mul.reshape([rshift_mul.shape[-1], 1]))
        return np.add.reduce(mul, axis=1)

    if mul_shift == 0 and rshift_mul_round.all() == 0 and rshift_mul.all() == 0:
        my_matmul = my_matmul_by_matmul
    else:
        my_matmul = my_matmul_by_multiply

    if act_func is None:
        def act_op(x): return x
    else:
        act_op = act_func.get_act_func()

    for bat in range(shape[0]):
        w = filter.reshape([shape[3], -1])

        oy = 0
        for py in range(-pad_row_top, input.shape[1] + pad_row_bottom, strides[1]):

            ox = 0
            for px in range(-pad_col_left, input.shape[2] + pad_col_right, strides[2]):

                ys = py + pad_row_top
                ye = ys + filter.shape[1]
                xs = px + pad_col_left
                xe = xs + filter.shape[2]
                a = input[bat, ys: ye, xs: xe].reshape([-1])

                sum = my_matmul(a, w)

                sum = np.left_shift(sum, sum_shift)
                sum = np.add(sum, rshift_sum_round)
                sum = np.right_shift(sum, rshift_sum)
                sum = np.add(sum, shifted_bias)
                sum = np.multiply(sum, shifted_scale)
                frac = np.where(rshift_out != 0, np.where(sum >= 0, rshift_out_round, rshift_out_round - 1),
                                np.zeros_like(rshift_out, dtype=np.int64))
                sum = np.add(sum, frac)
                sum = np.right_shift(sum, rshift_out)
                sum = np.where(sum > p_th, p_th, np.where(sum < n_th, n_th, sum))

                out[bat][oy][ox][:] = act_op(sum)

                ox += 1
                if ox >= out.shape[2]:
                    break

            oy += 1
            if oy >= out.shape[1]:
                break

    return out



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng
import veriloggen

import matrix_conv2d


act_shape = (1, 7, 7, 15)
weight_shape = (7, 3, 3, 15)
bias_shape = None
scale_shape = None
act_dtype = ng.int32
weight_dtype = ng.int32
bias_dtype = ng.int32
scale_dtype = ng.int32
out_dtype = ng.int32
stride = (1, 1, 1, 1)
rshift_mul = None
rshift_sum = None
rshift_out = None
act_func = None
par_ich = 1
par_och = 1
par_col = 1
par_row = 1
concur_och = None
stationary = 'filter'
input_ram_size = None
filter_ram_size = None
bias_ram_size = None
scale_ram_size = None
out_ram_size = None
axi_datawidth = 32


def test(request, silent=True):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    rslt = matrix_conv2d.run(act_shape, weight_shape,
                             bias_shape, scale_shape,
                             act_dtype, weight_dtype,
                             bias_dtype, scale_dtype,
                             out_dtype,
                             stride,
                             rshift_mul, rshift_sum, rshift_out,
                             act_func,
                             par_ich, par_och, par_col, par_row,
                             concur_och, stationary,
                             input_ram_size, filter_ram_size,
                             bias_ram_size, scale_ram_size,
                             out_ram_size,
                             axi_datawidth, silent,
                             filename=None, simtype=simtype,
                             outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    verify_rslt = [line for line in rslt.splitlines() if line.startswith('# verify:')][0]
    assert(verify_rslt == '# verify: PASSED')


if __name__ == '__main__':
    rslt = matrix_conv2d.run(act_shape, weight_shape,
                             bias_shape, scale_shape,
                             act_dtype, weight_dtype,
                             bias_dtype, scale_dtype,
                             out_dtype,
                             stride,
                             rshift_mul, rshift_sum, rshift_out,
                             act_func,
                             par_ich, par_och, par_col, par_row,
                             concur_och, stationary,
                             input_ram_size, filter_ram_size,
                             bias_ram_size, scale_ram_size,
                             out_ram_size,
                             axi_datawidth, silent=False,
                             filename='tmp.v',
                             outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.basic_types as bt


class placeholder(bt._Storage):

    def __init__(self, dtype, shape=None, name=None):
        bt._Storage.__init__(self,
                             dtype=dtype, shape=shape, name=name,
                             is_input=True)


class variable(bt._Storage):

    def __init__(self, dtype, shape=None, name=None):
        bt._Storage.__init__(self,
                             dtype=dtype, shape=shape, name=name,
                             is_input=False)


class constant(bt._Constant):

    def __init__(self, value, dtype=None, shape=None, name=None):
        bt._Constant.__init__(self, value,
                              dtype=dtype, shape=shape, name=name)


class zeros(constant):

    def __init__(self, shape, dtype=None, name=None):
        value = np.zeros(shape, dtype=np.int64)
        constant.__init__(self, value, dtype, shape, name)


def zeros_like(x, dtype=None, name=None):
    shape = x.shape
    return zeros(shape, dtype, name)


class ones(constant):

    def __init__(self, shape, dtype=None, name=None):
        value = np.ones(shape, dtype=np.int64)
        if dtype is not None and dtype.point > 0:
            value = value << [dtype.point]
        if dtype is not None and dtype.point < 0:
            value = value >> [-dtype.point]
        constant.__init__(self, value, dtype, shape, name)


def ones_like(x, dtype=None, name=None):
    shape = x.shape
    return ones(shape, dtype, name)


class full(constant):

    def __sub_str__(self):
        return ' fill_value:%d' % self.fill_value

    def __init__(self, shape, fill_value, dtype=None, name=None):
        value = np.full(shape, fill_value, dtype=np.int64)
        self.fill_value = fill_value
        if dtype is not None and dtype.point > 0:
            value = value << [dtype.point]
        if dtype is not None and dtype.point < 0:
            value = value >> [-dtype.point]
        constant.__init__(self, value, dtype, shape, name)


def full_like(x, fill_value, dtype=None, name=None):
    shape = x.shape
    return full(shape, fill_value, dtype, name)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.operator as operator

from . import util


def Slice(visitor, node):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    srcs = [util.optimize_to_raw_value(src) for src in srcs]

    input = srcs[0]
    if isinstance(input, (tuple, list)):
        input = np.array(input)

    shape = input.shape

    starts = [0 for s in shape]
    ends = [s for s in shape]
    axes = [i for i in range(len(shape))]
    steps = [1 for s in shape]

    # for Slice-1 (Deprecated)
    for attribute in node.attribute:
        if attribute.name == 'starts':
            starts = [v for v in attribute.ints]

        elif attribute.name == 'ends':
            ends = [v for v in attribute.ints]

        elif attribute.name == 'axes':
            axes = [v for v in attribute.ints]

    if len(srcs) > 1:
        starts = srcs[1]
        ends = srcs[2]
        axes = srcs[3]
        steps = srcs[4]

    if isinstance(input, (tuple, list, np.ndarray)):
        input = np.array(input)
        v = get_sliced_value(input, starts, ends, axes, steps)
        return v

    layout = input.get_layout()
    onnx_layout = input.get_onnx_layout()

    if layout is not None and onnx_layout is not None:
        axes = [layout.index(onnx_layout[axis]) for axis in axes]
        axes.sort()

    starts, ends, steps = extract_slices(input, starts, ends, axes, steps)
    return operator.slice_(input, starts, ends, steps)


def get_sliced_value(input, starts, ends, axes, steps):

    slices = to_slices(input, starts, ends, axes, steps)
    return input[slices]


def to_slices(input, starts, ends, axes, steps):
    slices = []
    index = 0

    for start, end, axis, step in sorted(zip(starts, ends, axes, steps),
                                         key=lambda x: x[2]):

        while index < axis:
            slices.append(slice(0, input.shape[index]))
            index += 1

        slices.append(slice(start, end, step))
        index += 1

    return tuple(slices)


def extract_slices(input, starts, ends, axes, steps):
    ret_starts = []
    ret_ends = []
    ret_steps = []
    index = 0

    for start, end, axis, step in sorted(zip(starts, ends, axes, steps),
                                         key=lambda x: x[2]):

        while index < axis:
            ret_starts.append(0)
            ret_ends.append(input.shape[index])
            ret_steps.append(1)
            index += 1

        ret_starts.append(start)
        ret_ends.append(end)
        ret_steps.append(step)
        index += 1

    return tuple(ret_starts), tuple(ret_ends), tuple(ret_steps)



"""
   NNgen: A Fully-Customizable Hardware Synthesis Compiler for Deep Neural Network

   Copyright 2017, Shinya Takamaeda-Yamazaki and Contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from __future__ import absolute_import
from __future__ import print_function

import sys
import os

import nngen as ng


# --------------------
# (1) Represent a DNN model as a dataflow by NNgen operators
# --------------------

# data types
act_dtype = ng.int8
weight_dtype = ng.int8
bias_dtype = ng.int32
scale_dtype = ng.int8
batchsize = 1

# input
input_layer = ng.placeholder(dtype=act_dtype,
                             shape=(batchsize, 32, 32, 3),  # N, H, W, C
                             name='input_layer')

# layer 0: conv2d (with bias and scale (= batchnorm)), relu, max_pool
w0 = ng.variable(dtype=weight_dtype,
                 shape=(64, 3, 3, 3),  # Och, Ky, Kx, Ich
                 name='w0')
b0 = ng.variable(dtype=bias_dtype,
                 shape=(w0.shape[0],), name='b0')
s0 = ng.variable(dtype=scale_dtype,
                 shape=(w0.shape[0],), name='s0')

a0 = ng.conv2d(input_layer, w0,
               strides=(1, 1, 1, 1),
               bias=b0,
               scale=s0,
               act_func=ng.relu,
               dtype=act_dtype,
               sum_dtype=ng.int32)

a0p = ng.max_pool_serial(a0,
                         ksize=(1, 2, 2, 1),
                         strides=(1, 2, 2, 1))

# layer 1: conv2d, relu, reshape
w1 = ng.variable(weight_dtype,
                 shape=(64, 3, 3, a0.shape[-1]),
                 name='w1')
b1 = ng.variable(bias_dtype,
                 shape=(w1.shape[0],),
                 name='b1')
s1 = ng.variable(scale_dtype,
                 shape=(w1.shape[0],),
                 name='s1')

a1 = ng.conv2d(a0p, w1,
               strides=(1, 1, 1, 1),
               bias=b1,
               scale=s1,
               act_func=ng.relu,
               dtype=act_dtype,
               sum_dtype=ng.int32)

a1r = ng.reshape(a1, [batchsize, -1])

# layer 2: full-connection, relu
w2 = ng.variable(weight_dtype,
                 shape=(256, a1r.shape[-1]),
                 name='w2')
b2 = ng.variable(bias_dtype,
                 shape=(w2.shape[0],),
                 name='b2')
s2 = ng.variable(scale_dtype,
                 shape=(w2.shape[0],),
                 name='s2')

a2 = ng.matmul(a1r, w2,
               bias=b2,
               scale=s2,
               transposed_b=True,
               act_func=ng.relu,
               dtype=act_dtype,
               sum_dtype=ng.int32)

# layer 3: full-connection, relu
w3 = ng.variable(weight_dtype,
                 shape=(10, a2.shape[-1]),
                 name='w3')
b3 = ng.variable(bias_dtype,
                 shape=(w3.shape[0],),
                 name='b3')
s3 = ng.variable(scale_dtype,
                 shape=(w3.shape[0],),
                 name='s3')

# output
output_layer = ng.matmul(a2, w3,
                         bias=b3,
                         scale=s3,
                         transposed_b=True,
                         name='output_layer',
                         dtype=act_dtype,
                         sum_dtype=ng.int32)


# --------------------
# (2) Assign weights to the NNgen operators
# --------------------

# In this example, random floating-point values are assigned.
# In a real case, you should assign actual weight values
# obtianed by a training on DNN framework.

# If you don't you NNgen's quantizer, you can assign integer weights to each tensor.


import numpy as np

w0_value = np.random.normal(size=w0.length).reshape(w0.shape)
w0_value = np.clip(w0_value, -3.0, 3.0)
w0.set_value(w0_value)

b0_value = np.random.normal(size=b0.length).reshape(b0.shape)
b0_value = np.clip(b0_value, -3.0, 3.0)
b0.set_value(b0_value)

s0_value = np.ones(s0.shape)
s0.set_value(s0_value)

w1_value = np.random.normal(size=w1.length).reshape(w1.shape)
w1_value = np.clip(w1_value, -3.0, 3.0)
w1.set_value(w1_value)

b1_value = np.random.normal(size=b1.length).reshape(b1.shape)
b1_value = np.clip(b1_value, -3.0, 3.0)
b1.set_value(b1_value)

s1_value = np.ones(s1.shape)
s1.set_value(s1_value)

w2_value = np.random.normal(size=w2.length).reshape(w2.shape)
w2_value = np.clip(w2_value, -3.0, 3.0)
w2.set_value(w2_value)

b2_value = np.random.normal(size=b2.length).reshape(b2.shape)
b2_value = np.clip(b2_value, -3.0, 3.0)
b2.set_value(b2_value)

s2_value = np.ones(s2.shape)
s2.set_value(s2_value)

w3_value = np.random.normal(size=w3.length).reshape(w3.shape)
w3_value = np.clip(w3_value, -3.0, 3.0)
w3.set_value(w3_value)

b3_value = np.random.normal(size=b3.length).reshape(b3.shape)
b3_value = np.clip(b3_value, -3.0, 3.0)
b3.set_value(b3_value)

s3_value = np.ones(s3.shape)
s3.set_value(s3_value)

# Quantizing the floating-point weights by the NNgen quantizer.
# Alternatively, you can assign integer weights by yourself to each tensor.

imagenet_mean = np.array([0.485, 0.456, 0.406]).astype(np.float32)
imagenet_std = np.array([0.229, 0.224, 0.225]).astype(np.float32)

if act_dtype.width > 8:
    act_scale_factor = 128
else:
    act_scale_factor = int(round(2 ** (act_dtype.width - 1) * 0.5))

input_scale_factors = {'input_layer': act_scale_factor}
input_means = {'input_layer': imagenet_mean * act_scale_factor}
input_stds = {'input_layer': imagenet_std * act_scale_factor}

ng.quantize([output_layer], input_scale_factors, input_means, input_stds)


# --------------------
# (3) Assign hardware attributes
# --------------------

# conv2d, matmul
# par_ich: parallelism in input-channel
# par_och: parallelism in output-channel
# par_col: parallelism in pixel column
# par_row: parallelism in pixel row

par_ich = 2
par_och = 2

a0.attribute(par_ich=par_ich, par_och=par_och)
a1.attribute(par_ich=par_ich, par_och=par_och)
a2.attribute(par_ich=par_ich, par_och=par_och)
output_layer.attribute(par_ich=par_ich, par_och=par_och)

# cshamt_out: right shift amount after applying bias/scale
# If you assign integer weights by yourself to each tensor,
# cshamt (constant shift amount) must be assigned to each operator.

# a0.attribute(cshamt_out=weight_dtype.width + 1)
# a1.attribute(cshamt_out=weight_dtype.width + 1)
# a2.attribute(cshamt_out=weight_dtype.width + 1)
# output_layer.attribute(cshamt_out=weight_dtype.width + 1)

# max_pool
# par: parallelism in in/out channel

par = par_och

a0p.attribute(par=par)


# --------------------
# (4) Verify the DNN model behavior by executing the NNgen dataflow as a software
# --------------------

# In this example, random integer values are assigned.
# In real case, you should assign actual integer activation values, such as an image.

input_layer_value = np.random.normal(size=input_layer.length).reshape(input_layer.shape)
input_layer_value = input_layer_value * imagenet_std + imagenet_mean
input_layer_value = np.clip(input_layer_value, -3.0, 3.0)
input_layer_value = input_layer_value * act_scale_factor
input_layer_value = np.clip(input_layer_value,
                            -1 * 2 ** (act_dtype.width - 1) - 1, 2 ** (act_dtype.width - 1))
input_layer_value = np.round(input_layer_value).astype(np.int64)

eval_outs = ng.eval([output_layer], input_layer=input_layer_value)
output_layer_value = eval_outs[0]

print(output_layer_value)


# --------------------
# (5) Convert the NNgen dataflow to a hardware description (Verilog HDL and IP-XACT)
# --------------------

silent = False
axi_datawidth = 32

# to Veriloggen object
# targ = ng.to_veriloggen([output_layer], 'hello_nngen', silent=silent,
#                        config={'maxi_datawidth': axi_datawidth})

# to IP-XACT (the method returns Veriloggen object, as well as to_veriloggen)
targ = ng.to_ipxact([output_layer], 'hello_nngen', silent=silent,
                    config={'maxi_datawidth': axi_datawidth})
print('# IP-XACT was generated. Check the current directory.')

# to Verilog HDL RTL (the method returns a source code text)
# rtl = ng.to_verilog([output_layer], 'hello_nngen', silent=silent,
#                    config={'maxi_datawidth': axi_datawidth})


# --------------------
# (6) Save the quantized weights
# --------------------

# convert weight values to a memory image:
# on a real FPGA platform, this image will be used as a part of the model definition.

param_filename = 'hello_nngen.npz'
chunk_size = 64

param_data = ng.export_ndarray([output_layer], chunk_size)
np.savez_compressed(param_filename, param_data)


# --------------------
# (7) Simulate the generated hardware by Veriloggen and Verilog simulator
# --------------------

# If you don't check the RTL behavior, exit here.
# print('# Skipping RTL simulation. If you simulate the RTL behavior, comment out the next line.')
# sys.exit()

import math
from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi

outputfile = 'hello_nngen.out'
filename = 'hello_nngen.v'
# simtype = 'iverilog'
simtype = 'verilator'

param_bytes = len(param_data)

variable_addr = int(
    math.ceil((input_layer.addr + input_layer.memory_size) / chunk_size)) * chunk_size
check_addr = int(math.ceil((variable_addr + param_bytes) / chunk_size)) * chunk_size
tmp_addr = int(math.ceil((check_addr + output_layer.memory_size) / chunk_size)) * chunk_size

memimg_datawidth = 32
mem = np.zeros([1024 * 1024 * 256 // memimg_datawidth], dtype=np.int64)
mem = mem + [100]

# placeholder
axi.set_memory(mem, input_layer_value, memimg_datawidth,
               act_dtype.width, input_layer.addr,
               max(int(math.ceil(axi_datawidth / act_dtype.width)), par_ich))

# parameters (variable and constant)
axi.set_memory(mem, param_data, memimg_datawidth,
               8, variable_addr)

# verification data
axi.set_memory(mem, output_layer_value, memimg_datawidth,
               act_dtype.width, check_addr,
               max(int(math.ceil(axi_datawidth / act_dtype.width)), par_och))

# test controller
m = Module('test')
params = m.copy_params(targ)
ports = m.copy_sim_ports(targ)
clk = ports['CLK']
resetn = ports['RESETN']
rst = m.Wire('RST')
rst.assign(Not(resetn))

# AXI memory model
if outputfile is None:
    outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

memimg_name = 'memimg_' + outputfile

memory = axi.AxiMemoryModel(m, 'memory', clk, rst,
                            datawidth=axi_datawidth,
                            memimg=mem, memimg_name=memimg_name,
                            memimg_datawidth=memimg_datawidth)
memory.connect(ports, 'maxi')

# AXI-Slave controller
_saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
_saxi.connect(ports, 'saxi')

# timer
time_counter = m.Reg('time_counter', 32, initval=0)
seq = Seq(m, 'seq', clk, rst)
seq(
    time_counter.inc()
)


def ctrl():
    for i in range(100):
        pass

    ng.sim.set_global_addrs(_saxi, tmp_addr)

    start_time = time_counter.value
    ng.sim.start(_saxi)

    print('# start')

    ng.sim.wait(_saxi)
    end_time = time_counter.value

    print('# end')
    print('# execution cycles: %d' % (end_time - start_time))

    # verify
    ok = True
    for bat in range(output_layer.shape[0]):
        for x in range(output_layer.shape[1]):
            orig = memory.read_word(bat * output_layer.aligned_shape[1] + x,
                                    output_layer.addr, act_dtype.width)
            check = memory.read_word(bat * output_layer.aligned_shape[1] + x,
                                     check_addr, act_dtype.width)

            if vthread.verilog.NotEql(orig, check):
                print('NG (', bat, x,
                      ') orig: ', orig, ' check: ', check)
                ok = False
            else:
                print('OK (', bat, x,
                      ') orig: ', orig, ' check: ', check)

    if ok:
        print('# verify: PASSED')
    else:
        print('# verify: FAILED')

    vthread.finish()


th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
fsm = th.start()

uut = m.Instance(targ, 'uut',
                 params=m.connect_params(targ),
                 ports=m.connect_ports(targ))

# simulation.setup_waveform(m, uut)
simulation.setup_clock(m, clk, hperiod=5)
init = simulation.setup_reset(m, resetn, m.make_reset(), period=100, polarity='low')

init.add(
    Delay(10000000),
    Systask('finish'),
)

# output source code
if filename is not None:
    m.to_verilog(filename)

# run simulation
sim = simulation.Simulator(m, sim=simtype)
rslt = sim.run(outputfile=outputfile)

print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


def sigmoid(visitor, node):
    features = node.args[0]

    visitor.visit(features)

    out_width = node.dtype.width
    out_point = node.dtype.point
    out_signed = node.dtype.signed
    if out_signed:
        out_scale = round((2 ** (out_width - 1)) * node.range_rate)
    else:
        out_scale = round((2 ** out_width) * node.range_rate)

    if out_point == 0:
        th_scale = out_scale
    elif out_point > 0:
        th_scale = out_scale >> out_point
    else:
        th_scale = out_scale << -1 * out_point

    node.scale_factor = float(th_scale)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
import functools

import nngen.basic_types as bt


class leaky_relu_base(bt._ActFuncOperator):
    input_chainable = True
    output_chainable = True
    slope = None
    rshift = None

    def __init__(self, features, dtype=None, name=None, par=1):
        shape = None
        bt._ActFuncOperator.__init__(self, features,
                                     dtype=dtype, shape=shape, name=name, par=par)

    @classmethod
    def get_eval_method(cls):
        import nngen.verify as verify

        name = 'leaky_relu'
        method = getattr(verify, name, None)
        method = functools.partial(method, slope=cls.slope, rshift=cls.rshift)
        return method

    def eval(self, memo, input_dict, **kwargs):
        if id(self) in memo:
            return memo[id(self)]

        import nngen.verify as verify

        name = 'leaky_relu'

        args = [arg.eval(memo, input_dict)
                for arg in self.args]

        kwargs['slope'] = self.slope
        kwargs['rshift'] = self.rshift
        kwargs['dtype'] = self.dtype
        kwargs['name'] = self.name
        kwargs['par'] = self.par

        method = self.get_eval_method()
        ret = method(*args, **kwargs)
        memo[id(self)] = ret

        return ret


leaky_relu_cache = {}


def get_leaky_relu_op(slope, rshift=None, dtype=None):

    if rshift is None:
        rshift = dtype.width - 1 if dtype is not None else 31

    if (slope, rshift) in leaky_relu_cache:
        return leaky_relu_cache[(slope, rshift)]

    @staticmethod
    def op(strm, *args, **kwargs):
        mul = strm.Mul(args[0], slope)
        slope_width = int(math.ceil(math.log(slope, 2)) + 1) + 1  # signed
        dtype_width = dtype.width if dtype is not None else 32
        mul.width = slope_width + dtype_width
        neg = strm.SraRound(mul, rshift)
        neg.width = dtype_width
        return strm.Mux(args[0] > strm.Int(0), args[0], neg)

    cls = type(
        'leaky_relu_%d_%d' % (slope, rshift),
        (leaky_relu_base,),
        {'op': op,
         'slope': slope,
         'rshift': rshift}
    )

    leaky_relu_cache[(slope, rshift)] = cls

    return cls


def leaky_relu(features, slope, rshift=None, dtype=None, name=None, par=1):

    op = get_leaky_relu_op(slope, rshift, dtype)

    return op(features, dtype, name, par)



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import functools
import math
import numpy as np

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def run(act_shape=(1, 7, 7, 15),
        act_dtype=ng.int32, out_dtype=ng.int32,
        begins=(0, 2, 2, 0), ends=(1, 6, 6, 3), strides=(1, 1, 1, 1),
        par=1,
        axi_datawidth=32, silent=False,
        filename=None, simtype='iverilog', outputfile=None):

    # create target hardware
    act = ng.placeholder(act_dtype, shape=act_shape, name='act')
    out = ng.slice_(act, begins, ends, strides,
                    dtype=out_dtype, name='out',
                    par=par)

    targ = ng.to_veriloggen([out], 'matrix_slice', silent=silent,
                            config={'maxi_datawidth': axi_datawidth})

    # verification data
    vact = np.arange(act.length, dtype=np.int64).reshape(act.shape)

    eval_outs = ng.eval([out], act=vact)
    vout = eval_outs[0]

    # to memory image
    size_max = int(math.ceil(max(act.memory_size, out.memory_size) / 4096)) * 4096
    check_addr = max(act.addr, out.addr) + size_max
    size_check = size_max
    tmp_addr = check_addr + size_check

    memimg_datawidth = 32
    mem = np.zeros([1024 * 1024 * 8 // (memimg_datawidth // 8)], dtype=np.int64)
    mem = mem + [100]

    axi.set_memory(mem, vact, memimg_datawidth,
                   act_dtype.width, act.addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), par))
    axi.set_memory(mem, vout, memimg_datawidth,
                   out_dtype.width, check_addr,
                   max(int(math.ceil(axi_datawidth / out_dtype.width)), par))

    # test controller
    m = Module('test')
    params = m.copy_params(targ)
    ports = m.copy_sim_ports(targ)
    clk = ports['CLK']
    resetn = ports['RESETN']
    rst = m.Wire('RST')
    rst.assign(Not(resetn))

   # AXI memory model
    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst,
                                datawidth=axi_datawidth,
                                memimg=mem, memimg_name=memimg_name,
                                memimg_datawidth=memimg_datawidth)
    memory.connect(ports, 'maxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    # timer
    time_counter = m.Reg('time_counter', 32, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        time_counter.inc()
    )

    def ctrl():
        for i in range(100):
            pass

        ng.sim.set_global_addrs(_saxi, tmp_addr)

        start_time = time_counter.value
        ng.sim.start(_saxi)

        print('# start')

        ng.sim.wait(_saxi)
        end_time = time_counter.value

        print('# end')
        print('# execution cycles: %d' % (end_time - start_time))

        # verify
        ok = True
        for bat in range(out.shape[0]):
            for y in range(out.shape[1]):
                for x in range(out.shape[2]):
                    for ch in range(out.shape[3]):
                        orig = memory.read_word(bat * out.aligned_shape[1] * out.aligned_shape[2] * out.aligned_shape[3] +
                                                y * out.aligned_shape[2] * out.aligned_shape[3] +
                                                x * out.aligned_shape[3] + ch,
                                                out.addr, out_dtype.width)
                        check = memory.read_word(bat * out.aligned_shape[1] * out.aligned_shape[2] * out.aligned_shape[3] +
                                                 y * out.aligned_shape[2] * out.aligned_shape[3] +
                                                 x * out.aligned_shape[3] + ch,
                                                 check_addr, out_dtype.width)

                        if vthread.verilog.NotEql(orig, check):
                            print('NG (', bat, y, x, ch,
                                  ') orig: ', orig, ' check: ', check)
                            ok = False
                        # else:
                        #    print('OK (', bat, y, x, ch,
                        #          ') orig: ', orig, ' check: ', check)

        if ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

        vthread.finish()

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(targ, 'uut',
                     params=m.connect_params(targ),
                     ports=m.connect_ports(targ))

    # simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, resetn, m.make_reset(), period=100, polarity='low')

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    # output source code
    if filename is not None:
        m.to_verilog(filename)

    # run simulation
    sim = simulation.Simulator(m, sim=simtype)
    rslt = sim.run(outputfile=outputfile)

    return rslt


if __name__ == '__main__':
    rslt = run(silent=False, filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import math
import numpy as np

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng
from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def run(act_dtype=ng.int8, weight_dtype=ng.int8,
        bias_dtype=ng.int32, scale_dtype=ng.int8,
        par_ich=2, par_och=2,
        chunk_size=64, axi_datawidth=32, silent=False,
        weight_filename='cnn.npz',
        verilog_filename=None,
        sim_filename=None,
        # simtype='iverilog',
        simtype='verilator',
        # simtype=None,  # no RTL simulation
        ):

    # --------------------
    # (1) Represent a DNN model as a dataflow by NNgen operators
    # --------------------

    # input
    input_layer = ng.placeholder(dtype=act_dtype,
                                 shape=(1, 32, 32, 3),  # N, H, W, C
                                 name='input_layer')

    # layer 0: conv2d (with bias and scale (= batchnorm)), relu, max_pool
    w0 = ng.variable(dtype=weight_dtype,
                     shape=(64, 3, 3, 3),  # Och, Ky, Kx, Ich
                     name='w0')
    b0 = ng.variable(dtype=bias_dtype,
                     shape=(w0.shape[0],), name='b0')
    s0 = ng.variable(dtype=scale_dtype,
                     shape=(w0.shape[0],), name='s0')

    a0 = ng.conv2d(input_layer, w0,
                   strides=(1, 1, 1, 1),
                   bias=b0,
                   scale=s0,
                   act_func=ng.relu,
                   dtype=act_dtype,
                   sum_dtype=ng.int32)

    a0p = ng.max_pool_serial(a0,
                             ksize=(1, 2, 2, 1),
                             strides=(1, 2, 2, 1))

    # layer 1: conv2d, relu, reshape
    w1 = ng.variable(weight_dtype,
                     shape=(64, 3, 3, a0.shape[-1]),
                     name='w1')
    b1 = ng.variable(bias_dtype,
                     shape=(w1.shape[0],),
                     name='b1')
    s1 = ng.variable(scale_dtype,
                     shape=(w1.shape[0],),
                     name='s1')

    a1 = ng.conv2d(a0p, w1,
                   strides=(1, 1, 1, 1),
                   bias=b1,
                   scale=s1,
                   act_func=ng.relu,
                   dtype=act_dtype,
                   sum_dtype=ng.int32)

    a1r = ng.reshape(a1, [1, -1])

    # layer 2: full-connection, relu
    w2 = ng.variable(weight_dtype,
                     shape=(256, a1r.shape[-1]),
                     name='w2')
    b2 = ng.variable(bias_dtype,
                     shape=(w2.shape[0],),
                     name='b2')
    s2 = ng.variable(scale_dtype,
                     shape=(w2.shape[0],),
                     name='s2')

    a2 = ng.matmul(a1r, w2,
                   bias=b2,
                   scale=s2,
                   transposed_b=True,
                   act_func=ng.relu,
                   dtype=act_dtype,
                   sum_dtype=ng.int32)

    # layer 3: full-connection, relu
    w3 = ng.variable(weight_dtype,
                     shape=(10, a2.shape[-1]),
                     name='w3')
    b3 = ng.variable(bias_dtype,
                     shape=(w3.shape[0],),
                     name='b3')
    s3 = ng.variable(scale_dtype,
                     shape=(w3.shape[0],),
                     name='s3')

    # output
    output_layer = ng.matmul(a2, w3,
                             bias=b3,
                             scale=s3,
                             transposed_b=True,
                             name='output_layer',
                             dtype=act_dtype,
                             sum_dtype=ng.int32)

    # --------------------
    # (2) Assign weights to the NNgen operators
    # --------------------

    # In this example, random floating-point values are assigned.
    # In a real case, you should assign actual weight values
    # obtianed by a training on DNN framework.

    # If you don't you NNgen's quantizer, you can assign integer weights to each tensor.

    w0_value = np.random.normal(size=w0.length).reshape(w0.shape)
    w0_value = np.clip(w0_value, -3.0, 3.0)
    w0.set_value(w0_value)

    b0_value = np.random.normal(size=b0.length).reshape(b0.shape)
    b0_value = np.clip(b0_value, -3.0, 3.0)
    b0.set_value(b0_value)

    s0_value = np.ones(s0.shape)
    s0.set_value(s0_value)

    w1_value = np.random.normal(size=w1.length).reshape(w1.shape)
    w1_value = np.clip(w1_value, -3.0, 3.0)
    w1.set_value(w1_value)

    b1_value = np.random.normal(size=b1.length).reshape(b1.shape)
    b1_value = np.clip(b1_value, -3.0, 3.0)
    b1.set_value(b1_value)

    s1_value = np.ones(s1.shape)
    s1.set_value(s1_value)

    w2_value = np.random.normal(size=w2.length).reshape(w2.shape)
    w2_value = np.clip(w2_value, -3.0, 3.0)
    w2.set_value(w2_value)

    b2_value = np.random.normal(size=b2.length).reshape(b2.shape)
    b2_value = np.clip(b2_value, -3.0, 3.0)
    b2.set_value(b2_value)

    s2_value = np.ones(s2.shape)
    s2.set_value(s2_value)

    w3_value = np.random.normal(size=w3.length).reshape(w3.shape)
    w3_value = np.clip(w3_value, -3.0, 3.0)
    w3.set_value(w3_value)

    b3_value = np.random.normal(size=b3.length).reshape(b3.shape)
    b3_value = np.clip(b3_value, -3.0, 3.0)
    b3.set_value(b3_value)

    s3_value = np.ones(s3.shape)
    s3.set_value(s3_value)

    # Quantizing the floating-point weights by the NNgen quantizer.
    # Alternatively, you can assign integer weights by yourself to each tensor.

    imagenet_mean = np.array([0.485, 0.456, 0.406]).astype(np.float32)
    imagenet_std = np.array([0.229, 0.224, 0.225]).astype(np.float32)

    if act_dtype.width > 8:
        act_scale_factor = 128
    else:
        act_scale_factor = int(round(2 ** (act_dtype.width - 1) * 0.5))

    input_scale_factors = {'input_layer': act_scale_factor}
    input_means = {'input_layer': imagenet_mean * act_scale_factor}
    input_stds = {'input_layer': imagenet_std * act_scale_factor}

    ng.quantize([output_layer], input_scale_factors, input_means, input_stds)

    # --------------------
    # (3) Assign hardware attributes
    # --------------------

    # conv2d, matmul
    # par_ich: parallelism in input-channel
    # par_och: parallelism in output-channel
    # par_col: parallelism in pixel column
    # par_row: parallelism in pixel row

    a0.attribute(par_ich=par_ich, par_och=par_och)
    a1.attribute(par_ich=par_ich, par_och=par_och)
    a2.attribute(par_ich=par_ich, par_och=par_och)
    output_layer.attribute(par_ich=par_ich, par_och=par_och)

    # cshamt_out: right shift amount after applying bias/scale
    # If you assign integer weights by yourself to each tensor,
    # cshamt (constant shift amount) must be assigned to each operator.

    # a0.attribute(cshamt_out=weight_dtype.width + 1)
    # a1.attribute(cshamt_out=weight_dtype.width + 1)
    # a2.attribute(cshamt_out=weight_dtype.width + 1)
    # output_layer.attribute(cshamt_out=weight_dtype.width + 1)

    # max_pool
    # par: parallelism in in/out channel

    par = par_och

    a0p.attribute(par=par)

    # --------------------
    # (4) Verify the DNN model behavior by executing the NNgen dataflow as a software
    # --------------------

    # In this example, random integer values are assigned.
    # In real case, you should assign actual integer activation values, such as an image.

    input_layer_value = np.random.normal(size=input_layer.length).reshape(input_layer.shape)
    input_layer_value = input_layer_value * imagenet_std + imagenet_mean
    input_layer_value = np.clip(input_layer_value, -5.0, 5.0)
    input_layer_value = input_layer_value * act_scale_factor
    input_layer_value = np.clip(input_layer_value,
                                -1 * 2 ** (act_dtype.width - 1) - 1, 2 ** (act_dtype.width - 1))
    input_layer_value = np.round(input_layer_value).astype(np.int64)

    eval_outs = ng.eval([output_layer], input_layer=input_layer_value)
    output_layer_value = eval_outs[0]

    # print(output_layer_value)
    # breakpoint()

    # --------------------
    # (5) Convert the NNgen dataflow to a hardware description (Verilog HDL and IP-XACT)
    # --------------------

    # to Veriloggen object
    # targ = ng.to_veriloggen([output_layer], 'cnn', silent=silent,
    #                        config={'maxi_datawidth': axi_datawidth})

    # to IP-XACT (the method returns Veriloggen object, as well as to_veriloggen)
    targ = ng.to_ipxact([output_layer], 'cnn', silent=silent,
                        config={'maxi_datawidth': axi_datawidth})

    # to Verilog HDL RTL (the method returns a source code text)
    # rtl = ng.to_verilog([output_layer], 'cnn', silent=silent,
    #                    config={'maxi_datawidth': axi_datawidth})

    # --------------------
    # (6) Save the quantized weights
    # --------------------

    # convert weight values to a memory image:
    # on a real FPGA platform, this image will be used as a part of the model definition.

    param_filename = 'hello_nngen.npy'
    chunk_size = 64

    param_data = ng.export_ndarray([output_layer], chunk_size)
    np.savez_compressed(weight_filename, param_data)

    # --------------------
    # (7) Simulate the generated hardware by Veriloggen and Verilog simulator
    # --------------------

    if simtype is None:
        sys.exit()

    param_bytes = len(param_data)

    variable_addr = int(
        math.ceil((input_layer.addr + input_layer.memory_size) / chunk_size)) * chunk_size
    check_addr = int(math.ceil((variable_addr + param_bytes) / chunk_size)) * chunk_size
    tmp_addr = int(math.ceil((check_addr + output_layer.memory_size) / chunk_size)) * chunk_size

    memimg_datawidth = 32
    mem = np.zeros([1024 * 1024 * 256 // (memimg_datawidth // 8)], dtype=np.int64)
    mem = mem + [100]

    # placeholder
    axi.set_memory(mem, input_layer_value, memimg_datawidth,
                   act_dtype.width, input_layer.addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), par_ich))

    # parameters (variable and constant)
    axi.set_memory(mem, param_data, memimg_datawidth,
                   8, variable_addr)

    # verification data
    axi.set_memory(mem, output_layer_value, memimg_datawidth,
                   act_dtype.width, check_addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), par_och))

    # test controller
    m = Module('test')
    params = m.copy_params(targ)
    ports = m.copy_sim_ports(targ)
    clk = ports['CLK']
    resetn = ports['RESETN']
    rst = m.Wire('RST')
    rst.assign(Not(resetn))

    # AXI memory model
    if sim_filename is None:
        sim_filename = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + sim_filename

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst,
                                datawidth=axi_datawidth,
                                memimg=mem, memimg_name=memimg_name,
                                memimg_datawidth=memimg_datawidth)
    memory.connect(ports, 'maxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    # timer
    time_counter = m.Reg('time_counter', 32, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        time_counter.inc()
    )

    def ctrl():
        for i in range(100):
            pass

        ng.sim.set_global_addrs(_saxi, tmp_addr)

        start_time = time_counter.value
        ng.sim.start(_saxi)

        print('# start')

        ng.sim.wait(_saxi)
        end_time = time_counter.value

        print('# end')
        print('# execution cycles: %d' % (end_time - start_time))

        # verify
        ok = True
        for bat in range(output_layer.shape[0]):
            for x in range(output_layer.shape[1]):
                orig = memory.read_word(bat * output_layer.aligned_shape[1] + x,
                                        output_layer.addr, act_dtype.width)
                check = memory.read_word(bat * output_layer.aligned_shape[1] + x,
                                         check_addr, act_dtype.width)

                if vthread.verilog.NotEql(orig, check):
                    print('NG (', bat, x,
                          ') orig: ', orig, ' check: ', check)
                    ok = False
                else:
                    print('OK (', bat, x,
                          ') orig: ', orig, ' check: ', check)

        if ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

        vthread.finish()

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(targ, 'uut',
                     params=m.connect_params(targ),
                     ports=m.connect_ports(targ))

    # simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, resetn, m.make_reset(), period=100, polarity='low')

    init.add(
        Delay(10000000),
        Systask('finish'),
    )

    # output source code
    if verilog_filename is not None:
        m.to_verilog(verilog_filename)

    # run simulation
    sim = simulation.Simulator(m, sim=simtype)
    rslt = sim.run(outputfile=sim_filename)

    return rslt


if __name__ == '__main__':
    rslt = run(silent=False, verilog_filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections

import nngen.storage as storage
import nngen.dtype_list as dtype_list

from . import util
from . import basic
from . import exp
from . import reduce
from . import conv
from . import gemm
from . import pool
from . import pad
from . import act_func
from . import batchnormalization
from . import shape
from . import reshape
from . import flatten
from . import upsample
from . import transpose
from . import concat
from . import squeeze
from . import gather
from . import slice_
from . import cast
from . import ceil
from . import floor
from . import identity


# describe custom ONNX converting methods here
func_map = {
    'Add': basic.Add,
    'Sub': basic.Sub,
    'Mul': basic.Mul,
    'Div': basic.Div,
    'Exp': exp.Exp,
    'ReduceSum': reduce.ReduceSum,
    'ReduceMax': reduce.ReduceMax,
    'ReduceMin': reduce.ReduceMin,
    'ArgMax': reduce.ArgMax,
    'ArgMin': reduce.ArgMin,
    'Conv': conv.Conv,
    'Gemm': gemm.Gemm,
    'AveragePool': pool.AveragePool,
    'GlobalAveragePool': pool.GlobalAveragePool,
    'MaxPool': pool.MaxPool,
    'Pad': pad.Pad,
    'Relu': act_func.Relu,
    'LeakyRelu': act_func.LeakyRelu,
    'Sigmoid': act_func.Sigmoid,
    'BatchNormalization': batchnormalization.BatchNormalization,
    'Shape': shape.Shape,
    'Reshape': reshape.Reshape,
    'Flatten': flatten.Flatten,
    'Upsample': upsample.Upsample,
    'Transpose': transpose.Transpose,
    'Concat': concat.Concat,
    'Squeeze': squeeze.Squeeze,
    'Unsqueeze': squeeze.Unsqueeze,
    'Gather': gather.Gather,
    'Slice': slice_.Slice,
    'Cast': cast.Cast,
    'Ceil': ceil.Ceil,
    'Floor': floor.Floor,
    'Identity': identity.Identity,
}


def _get_func(op_type):
    return func_map[op_type]


class _OperatorVisitor(object):

    def __init__(self, model,
                 placeholders, variables, constants, operators,
                 producers, consumers,
                 value_dtypes,
                 default_placeholder_dtype, default_variable_dtype,
                 default_constant_dtype, default_operator_dtype,
                 default_scale_dtype, default_bias_dtype,
                 onnx_input_layout=('N', 'C', 'H', 'W'),
                 onnx_filter_layout=('O', 'I', 'H', 'W'),
                 nngen_input_layout=('N', 'H', 'W', 'C'),
                 nngen_filter_layout=('O', 'H', 'W', 'I'),
                 disable_fusion=False, verbose=False):

        self.model = model

        self.placeholders = placeholders
        self.variables = variables
        self.constants = constants
        self.operators = operators

        self.producers = producers
        self.consumers = consumers

        self.value_dtypes = value_dtypes

        self.default_placeholder_dtype = default_placeholder_dtype
        self.default_variable_dtype = default_variable_dtype
        self.default_constant_dtype = default_constant_dtype
        self.default_operator_dtype = default_operator_dtype
        self.default_scale_dtype = default_scale_dtype
        self.default_bias_dtype = default_bias_dtype

        self.onnx_input_layout = onnx_input_layout
        self.onnx_filter_layout = onnx_filter_layout

        self.nngen_input_layout = nngen_input_layout
        self.nngen_filter_layout = nngen_filter_layout

        self.disable_fusion = disable_fusion

        self.verbose = verbose

    def _verbose_node(self, name, node_func, ret):
        if self.verbose:
            print('[onnx] {} {} -> {} {}'.format(name, node_func, type(ret), str(ret)))
        return ret

    def visit(self, name):
        if name in self.placeholders:
            # return self.placeholders[name]
            return self._verbose_node(name, None, self.placeholders[name])

        if name in self.variables:
            # return self.variables[name]
            return self._verbose_node(name, None, self.variables[name])

        if name in self.constants:
            # return self.constants[name]
            return self._verbose_node(name, None, self.constants[name])

        if name in self.operators:
            # return self.operators[name]
            return self._verbose_node(name, None, self.operators[name])

        node = util.search_node_from_model(self.model, name)

        node_func = _get_func(node.op_type)
        node_op = node_func(self, node)

        output_names = util.get_output_names(node)
        for output_name in output_names:
            self.operators[output_name] = node_op

        # return node_op
        return self._verbose_node(name, node_func, node_op)


def from_onnx(filename,
              value_dtypes=None,
              value_shapes=None,
              default_placeholder_dtype=dtype_list.int32,
              default_variable_dtype=dtype_list.int32,
              default_constant_dtype=dtype_list.int32,
              default_operator_dtype=dtype_list.int32,
              default_scale_dtype=dtype_list.int32,
              default_bias_dtype=dtype_list.int32,
              onnx_input_layout=('N', 'C', 'H', 'W'),
              onnx_filter_layout=('O', 'I', 'H', 'W'),
              disable_fusion=False, verbose=False):
    """
    Convert ONNX model to NNgen model

    Parameters
    ----------
    filename : str
        File name of ONNX model

    value_dtypes : dict
        dtype_info dictionary by name

    value_shapes : dict
        shape dictionary for undefined node shapes by name

    default_placeholder_dtype : nngen.dtype_info
        Default dtype for placeholder

    default_variable_dtype : nngen.dtype_info
        Default dtype for variable

    default_constant_dtype : nngen.dtype_info
        Default dtype for constant

    default_operator_dtype : nngen.dtype_info
        Default dtype for operator

    default_scale_dtype : nngen.dtype_info
        Default dtype for scale

    default_bias_dtype : nngen.dtype_info
        Default dtype for bias

    onnx_input_layout : str
        Layout of ONNX input values

    onnx_filter_layout : str
        Layout of ONNX filter (weight) values

    disable_fusion : bool
        Disable operator fusion

    Returns
    -------
    outputs : collections.OrderedDict
        Dict of output values

    placeholders : collections.OrderedDict
        Dictionary of placeholders

    variables : collections.OrderedDict
        Dictionary of variables

    constants : collections.OrderedDict
        Dictionary of constants

    operators : collections.OrderedDict
        Dictionary of operators
    """

    try:
        import onnx
        from onnx import numpy_helper
    except:
        raise ImportError('onnx is required.')

    if value_dtypes is None:
        value_dtypes = {}

    if value_shapes is None:
        value_shapes = {}

    # load model
    model = onnx.load(filename)

    # input/output node dict
    input_nodes = collections.OrderedDict()
    output_nodes = collections.OrderedDict()

    for input_var in model.graph.input:
        input_nodes[input_var.name] = input_var

    for output_var in model.graph.output:
        output_nodes[output_var.name] = output_var

    # variable ndarray dict
    variable_values = collections.OrderedDict()

    for weight in model.graph.initializer:
        name = weight.name
        np_weight = numpy_helper.to_array(weight)
        variable_values[name] = np_weight

    # constant ndarray dict
    constant_values = collections.OrderedDict()

    for node in model.graph.node:
        if node.op_type == 'Constant':
            value = numpy_helper.to_array(node.attribute[0].t)
            output_names = util.get_output_names(node)
            for output_name in output_names:
                constant_values[output_name] = value

    # placeholders
    placeholders = _to_placeholders(input_nodes, output_nodes,
                                    variable_values, constant_values,
                                    value_dtypes, value_shapes,
                                    default_placeholder_dtype,
                                    default_variable_dtype,
                                    default_constant_dtype,
                                    default_operator_dtype)

    # variables
    variables = _to_variables(input_nodes, output_nodes,
                              variable_values, constant_values,
                              value_dtypes, value_shapes,
                              default_placeholder_dtype,
                              default_variable_dtype,
                              default_constant_dtype,
                              default_operator_dtype)

    # constants
    constants = _to_constants(input_nodes, output_nodes,
                              variable_values, constant_values,
                              value_dtypes, value_shapes,
                              default_placeholder_dtype,
                              default_variable_dtype,
                              default_constant_dtype,
                              default_operator_dtype)

    # producer/consumer table
    producers = collections.defaultdict(list)
    consumers = collections.defaultdict(list)

    for node in model.graph.node:
        output_names = util.get_output_names(node)
        for output_name in output_names:
            for arg in node.input:
                if arg not in producers[output_name]:
                    producers[output_name].append(arg)
                if output_name not in consumers[arg]:
                    consumers[arg].append(output_name)

    # operators
    operators = collections.OrderedDict()
    visitor = _OperatorVisitor(model,
                               placeholders, variables, constants, operators,
                               producers, consumers,
                               value_dtypes,
                               default_placeholder_dtype, default_variable_dtype,
                               default_constant_dtype, default_operator_dtype,
                               default_scale_dtype, default_bias_dtype,
                               onnx_input_layout, onnx_filter_layout,
                               disable_fusion=disable_fusion, verbose=verbose)

    placeholders = visitor.placeholders
    variables = visitor.variables
    constants = visitor.constants
    operators = visitor.operators

    for name, output_node in output_nodes.items():
        visitor.visit(name)

    # outputs
    outputs = collections.OrderedDict()

    for name, node in output_nodes.items():
        if name in operators:
            outputs[name] = operators[name]
        elif name in placeholders:
            outputs[name] = placeholders[name]
        elif name in variables:
            outputs[name] = variables[name]
        elif name in constants:
            outputs[name] = constants[name]

    return outputs, placeholders, variables, constants, operators


def _to_placeholders(input_nodes, output_nodes, variable_values, constant_values,
                     value_dtypes, value_shapes,
                     default_placeholder_dtype, default_variable_dtype,
                     default_constant_dtype, default_operator_dtype):

    placeholders = collections.OrderedDict()

    for name, node in input_nodes.items():
        # exclude variables
        if name in variable_values:
            continue

        if name in value_dtypes:
            dtype = value_dtypes[name]
        else:
            dtype = default_placeholder_dtype

        shape = util.to_shape(node, value_shapes)
        p = storage.placeholder(dtype=dtype, shape=shape, name=name)
        placeholders[name] = p

    return placeholders


def _to_variables(input_nodes, output_nodes, variable_values, constant_values,
                  value_dtypes, value_shapes,
                  default_placeholder_dtype, default_variable_dtype,
                  default_constant_dtype, default_operator_dtype):

    variables = collections.OrderedDict()

    for name, node in variable_values.items():
        if name in value_dtypes:
            dtype = value_dtypes[name]
        else:
            dtype = default_variable_dtype

        shape = node.shape
        v = storage.variable(dtype=dtype, shape=shape, name=name)
        v.set_value(node)
        variables[name] = v

    return variables


def _to_constants(input_nodes, output_nodes, variable_values, constant_values,
                  value_dtypes, value_shapes,
                  default_placeholder_dtype, default_variable_dtype,
                  default_constant_dtype, default_operator_dtype):

    constants = collections.OrderedDict()

    for name, node in constant_values.items():
        if name in value_dtypes:
            dtype = value_dtypes[name]
        else:
            dtype = default_constant_dtype

        shape = node.shape
        c = storage.constant(value=node,
                             dtype=dtype, shape=shape, name=name)
        constants[name] = c

    return constants



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.operator as operator

from . import util
from . import reshape


def Flatten(visitor, node, no_transpose=False):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    input = srcs[0]

    for attribute in node.attribute:
        if attribute.name == 'axis':
            axis = attribute.i

    if (not no_transpose and
        input.get_layout() is not None and input.get_onnx_layout() is not None and
            input.get_layout() != input.get_onnx_layout()):

        perm = [input.get_layout().index(l) for l in input.get_onnx_layout()]
        onnx_perm = [i for i, l in enumerate(input.get_onnx_layout())]

        input = operator.transpose(input, perm)
        input.implicit = True
        input.transpose_onnx_perm = onnx_perm

        input.layout = input.get_onnx_layout()

    # no transpose is required below.
    layout = input.get_layout()
    onnx_layout = input.get_onnx_layout()

    input_shape = input.shape
    if layout is not None and onnx_layout is not None:
        input_shape = [input_shape[layout.index(ol)] for ol in onnx_layout]

    if axis == 0:
        shape = [1, -1]
    else:
        shape = [int(np.prod(input_shape[:axis])), -1]

    if layout is not None and onnx_layout is not None:
        length = input.get_length()
        shape = util.infer_shape(shape, length)

        out_onnx_layout = reshape.make_layout(input_shape, shape, onnx_layout)
        layout_map = reshape.layout_matching(input_shape, shape, onnx_layout, out_onnx_layout)
        layout_map = dict([(tuple(sorted(k)), v) for k, v in layout_map.items()])

        out_layout = []
        key_list = []
        for l in layout:
            key_list.append(l)
            key_list.sort()
            key = tuple(key_list)
            if key in layout_map:
                out_layout.extend(layout_map[key])
                key_list = []

        out_layout = tuple(out_layout)

        shape = [shape[out_onnx_layout.index(ol)]
                 for ol in out_layout]

    name = util.get_name(node)

    c = operator.reshape(input, shape, name=name)

    return c



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import functools
import math
import numpy as np

import torch
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as F
import torch.autograd

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


class MatrixSigmoid(nn.Module):
    def __init__(self):
        super(MatrixSigmoid, self).__init__()
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        z = self.sigmoid(x)
        return z


def run(act_shape=(15, 31),
        act_dtype=ng.int32,
        par=1,
        chunk_size=64,
        axi_datawidth=32, silent=False,
        filename=None, simtype='iverilog', outputfile=None):

    # pytorch model
    model = MatrixSigmoid()

    # Pytorch to ONNX
    onnx_filename = 'onnx_matrix_sigmoid.onnx'
    dummy_input = torch.randn(*act_shape)
    input_names = ['act']
    output_names = ['out']
    model.eval()
    torch.onnx.export(model, dummy_input, onnx_filename,
                      input_names=input_names, output_names=output_names)

    # --------------------
    # (1) Represent a DNN model as a dataflow by NNgen operators
    # --------------------

    # ONNX to NNgen
    dtypes = {'act': act_dtype,
              'out': act_dtype}

    (outputs, placeholders, variables,
     constants, operators) = ng.from_onnx(onnx_filename,
                                          value_dtypes=dtypes,
                                          default_placeholder_dtype=act_dtype,
                                          default_variable_dtype=ng.int32,
                                          default_constant_dtype=ng.int32,
                                          default_operator_dtype=act_dtype,
                                          default_scale_dtype=ng.int32,
                                          default_bias_dtype=ng.int32,
                                          disable_fusion=False)

    # --------------------
    # (2) Assign quantized weights to the NNgen operators
    # --------------------

    if act_dtype.width > 8:
        act_scale_factor = 128
    else:
        act_scale_factor = int(round(2 ** (act_dtype.width - 1) * 0.5))

    input_scale_factors = {'act': act_scale_factor}

    ng.quantize(outputs, input_scale_factors)

    # --------------------
    # (3) Assign hardware attributes
    # --------------------

    for op in operators.values():
        if isinstance(op, ng.sigmoid):
            op.attribute(par=par)

    # --------------------
    # (4) Verify the DNN model behavior by executing the NNgen dataflow as a software
    # --------------------

    act = placeholders['act']
    out = outputs['out']

    # verification data
    # random data
    #std = 0.2
    #mean = 0.5
    std = 3.0
    mean = 0.0
    img = np.random.normal(size=act.length).astype(np.float32).reshape(act.shape)
    img = img * std + mean

    # execution on pytorch
    model_input = img

    if act.perm is not None:
        model_input = np.transpose(model_input, act.reversed_perm)

    model.eval()
    model_out = model(torch.from_numpy(model_input)).detach().numpy()
    if act.perm is not None and len(model_out.shape) == len(act.shape):
        model_out = np.transpose(model_out, act.perm)
    scaled_model_out = model_out * out.scale_factor

    # software-based verification
    vact = img * act_scale_factor
    vact = np.clip(vact,
                   -1.0 * (2 ** (act.dtype.width - 1) - 1),
                   1.0 * (2 ** (act.dtype.width - 1) - 1))
    vact = np.round(vact).astype(np.int64)

    eval_outs = ng.eval([out], act=vact)
    vout = eval_outs[0]

    mean_square_error = np.sum((vout - scaled_model_out) ** 2) / vout.size
    corrcoef = np.corrcoef(model_out.reshape([-1]), vout.reshape([-1]))

    # breakpoint()

    # --------------------
    # (5) Convert the NNgen dataflow to a hardware description (Verilog HDL and IP-XACT)
    # --------------------

    targ = ng.to_veriloggen([out], 'onnx_matrix_sigmoid', silent=silent,
                            config={'maxi_datawidth': axi_datawidth})

    # --------------------
    # (6) Simulate the generated hardware by Veriloggen and Verilog simulator
    # --------------------

    if simtype is None:
        sys.exit()

    # to memory image
    param_data = ng.export_ndarray([out])
    param_bytes = len(param_data)

    variable_addr = int(math.ceil((act.addr + act.memory_size) / chunk_size)) * chunk_size
    check_addr = int(math.ceil((variable_addr + param_bytes) / chunk_size)) * chunk_size
    tmp_addr = int(math.ceil((check_addr + out.memory_size) / chunk_size)) * chunk_size

    memimg_datawidth = 32
    mem = np.zeros([1024 * 1024 * 8 // (memimg_datawidth // 8)], dtype=np.int64)
    mem = mem + [100]

    # placeholder
    axi.set_memory(mem, vact, memimg_datawidth,
                   act_dtype.width, act.addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), par))

    # parameters (variable and constant)
    axi.set_memory(mem, param_data, memimg_datawidth,
                   8, variable_addr)

    # verification data
    axi.set_memory(mem, vout, memimg_datawidth,
                   act_dtype.width, check_addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), par))

    # test controller
    m = Module('test')
    params = m.copy_params(targ)
    ports = m.copy_sim_ports(targ)
    clk = ports['CLK']
    resetn = ports['RESETN']
    rst = m.Wire('RST')
    rst.assign(Not(resetn))

    # AXI memory model
    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst,
                                datawidth=axi_datawidth,
                                memimg=mem, memimg_name=memimg_name,
                                memimg_datawidth=memimg_datawidth)
    memory.connect(ports, 'maxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    # timer
    time_counter = m.Reg('time_counter', 32, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        time_counter.inc()
    )

    num_rep = functools.reduce(lambda x, y: x * y, out.shape[:-1], 1)

    def ctrl():
        for i in range(100):
            pass

        ng.sim.set_global_addrs(_saxi, tmp_addr)

        start_time = time_counter.value
        ng.sim.start(_saxi)

        print('# start')

        ng.sim.wait(_saxi)
        end_time = time_counter.value

        print('# end')
        print('# execution cycles: %d' % (end_time - start_time))

        # verify
        ok = True
        for i in range(num_rep):
            for j in range(out.shape[-1]):
                orig = memory.read_word(i * out.aligned_shape[-1] + j,
                                        out.addr, act_dtype.width)
                check = memory.read_word(i * out.aligned_shape[-1] + j,
                                         check_addr, act_dtype.width)

                if vthread.verilog.NotEql(orig, check):
                    print('NG', i, j, orig, check)
                    ok = False
                # else:
                #    print('OK', i, j, orig, check)

        if ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

        vthread.finish()

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(targ, 'uut',
                     params=m.connect_params(targ),
                     ports=m.connect_ports(targ))

    # simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, resetn, m.make_reset(), period=100, polarity='low')

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    # output source code
    if filename is not None:
        m.to_verilog(filename)

    # run simulation
    sim = simulation.Simulator(m, sim=simtype)
    rslt = sim.run(outputfile=outputfile)

    return rslt


if __name__ == '__main__':
    rslt = run(silent=False, filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections

import nngen.operator as operator

from . import util


def Transpose(visitor, node):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    input = srcs[0]

    perm = [i for i in range(len(input.shape))]
    for attribute in node.attribute:
        if attribute.name == 'perm':
            perm = [i for i in attribute.ints]

    layout = input.get_layout()
    onnx_layout = input.get_onnx_layout()

    transposed_layout = None
    transposed_onnx_layout = None

    onnx_perm = perm

    if layout is not None and onnx_layout is not None:
        perm = util.convert_transpose_perm(perm, onnx_layout, layout)

        transposed_layout = tuple([layout[p] for p in perm])
        transposed_onnx_layout = tuple([onnx_layout[p] for p in onnx_perm])

        if layout == transposed_layout:
            v = operator.cast(input, input.dtype)
            v.layout = input.get_layout()
            v.onnx_layout = input.get_layout()
            return v

    kwargs = collections.OrderedDict()

    name = util.get_name(node)
    kwargs['name'] = name

    c = operator.transpose(input, perm, **kwargs)
    c.transpose_onnx_perm = onnx_perm

    if layout is not None and onnx_layout is not None:
        c.layout = transposed_layout
        c.onnx_layout = transposed_onnx_layout

    return c



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections
import numpy as np
import functools

import nngen.storage as storage
import nngen.operator as operator

from . import util


def _normalize_elementwise(method, pre_methods, visitor, node, np_method=None):

    srcs = []

    for src in list(node.input):
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    opt_srcs = [util.optimize_to_raw_value(src) for src in srcs]

    all_ndarray = True
    for src in opt_srcs:
        if not isinstance(src, np.ndarray):
            all_ndarray = False
            break

    if all_ndarray and np_method is not None:
        return np_method(*opt_srcs)

    name = util.get_name(node)

    if pre_methods is None:
        pre_methods = [None for _ in srcs]

    if len(pre_methods) != len(srcs):
        raise ValueError('length mismatch: %d != %d' % (len(pre_methods), len(srcs)))

    args = []
    for i, (src, pre_method) in enumerate(zip(srcs, pre_methods)):
        if pre_method is None:
            args.append(src)
        else:
            args.append(pre_method(src))

    for src in srcs:
        scale = 1
        args.append(scale)

    return method(*args)


def _elementwise(method, visitor, node, np_method=None):

    args = []

    for src in list(node.input):
        src_obj = visitor.visit(src)
        args.append(src_obj)

    opt_args = [util.optimize_to_raw_value(arg) for arg in args]

    all_ndarray = True
    for arg in opt_args:
        if not isinstance(arg, np.ndarray):
            all_ndarray = False
            break

    if all_ndarray and np_method is not None:
        return np_method(*opt_args)

    kwargs = collections.OrderedDict()

    name = util.get_name(node)
    kwargs['name'] = name

    if name in visitor.value_dtypes:
        kwargs['dtype'] = visitor.value_dtypes[name]

    return method(*args, **kwargs)


def Add(visitor, node):

    method = functools.partial(operator.scaled_add, shamt=0)
    return _normalize_elementwise(method, None, visitor, node, np.add)


def Sub(visitor, node):

    method = functools.partial(operator.scaled_add, shamt=0)
    pre_methods = (None, operator.neg)
    return _normalize_elementwise(method, pre_methods, visitor, node, np.subtract)


def Mul(visitor, node):

    method = functools.partial(operator.scaled_multiply, shamt=0)
    return _elementwise(method, visitor, node, np.multiply)


def Div(visitor, node):

    method = functools.partial(operator.scaled_div, shamt=0)
    return _elementwise(method, visitor, node, np.divide)



from setuptools import setup, find_packages

import os


def read(filename):
    # return open(os.path.join(os.path.dirname(__file__), filename), encoding='utf8').read()
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


setup(name='nngen',
      version=read('nngen/VERSION').splitlines()[0],
      description='A Fully-Customizable Hardware Synthesis Compiler for Deep Neural Network',
      long_description=read('README.md'),
      long_description_content_type="text/markdown",
      keywords='Neural Network, Deep Learning, FPGA, High-Level Synthesis',
      author='Shinya Takamaeda-Yamazaki',
      license="Apache License 2.0",
      url='https://github.com/NNgen/nngen',
      packages=find_packages(),
      package_data={'nngen': ['VERSION'], },
      install_requires=['veriloggen>=2.3.0',
                        'numpy>=1.17',
                        'onnx>=1.9.0'],
      extras_require={
          'test': ['pytest>=3.8.1', 'pytest-pythonpath>=0.7.3'],
      },
      )



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.operator as operator

from . import basic


def Exp(visitor, node):

    return basic._elementwise(operator.exp, visitor, node, np.exp)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
import functools
from decimal import Decimal, ROUND_HALF_UP, ROUND_HALF_EVEN


def leaky_relu(features, slope, rshift, dtype=None, name=None, par=1,
               features_dtype=None):

    if rshift is None:
        rshift = dtype.width if dtype is not None else 31

    features_point = 0 if features_dtype is None else features_dtype.point
    out_point = 0 if dtype is None else dtype.point
    out_shift = out_point - features_point

    features_shape = features.shape
    negs = np.array(list(map(lambda x: Decimal(str(x * slope / (2**rshift))).quantize(Decimal('0'),
                                                                                      rounding=ROUND_HALF_UP), features.flatten()))).astype(np.int64).reshape(features_shape)

    comp = features >= 0

    out_op = ((lambda x: x << out_shift) if out_shift >= 0 else
              (lambda x: x >> -out_shift))

    ret = out_op(np.where(comp, features, negs))

    return ret


def get_leaky_relu_op(slope, rshift=None, dtype=None):
    return functools.partial(leaky_relu,
                             slope=slope, rshift=rshift, dtype=dtype)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np


def extern(values, opcode,
           shape=None, dtype=None, name=None,
           value_dtypes=None, func=None):

    if not isinstance(values, (tuple, list)):
        values = [values]

    if not isinstance(opcode, int):
        raise TypeError("opcode must be int, not '%s'." %
                        str(type(opcode)))

    if shape is None:
        shape = values[0].shape

    if dtype is None:
        dtype = values[0].dtype

    if func is None:
        return np.ones(shape, dtype=np.int64)

    return func(*values)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
import math


def quantize_linear(orig_weight, width=8):
    if width > 16:
        width = 16

    return _quantize_linear(orig_weight, width)


def _quantize_linear(orig_weight, width=8, range_rate=0.995):
    if isinstance(orig_weight, (tuple, list)):
        orig_weight = np.array(orig_weight)

    if width >= 8:
        max_value = np.max(orig_weight)
        min_value = np.min(orig_weight)
    else:
        sorted_abs_weight = np.sort(np.abs(orig_weight).reshape([-1]))
        max_value = sorted_abs_weight[math.ceil(orig_weight.size * range_rate) - 1]
        min_value = - max_value
        orig_weight = np.clip(orig_weight, min_value, max_value)

    abs_max = max(abs(max_value), abs(min_value))

    pos_num_quantized_bins = 2 ** (width - 1) - 1
    scale_factor = 1.0 * pos_num_quantized_bins / abs_max
    quantized_weight = np.round(orig_weight * scale_factor).astype(np.int64)

    return quantized_weight, scale_factor


def quantize_linear_by_scale_factor(value, width, scale_factor):
    half_range = 2 ** (width - 1) - 1
    v = np.round(value * scale_factor).astype(np.int64)
    v = np.clip(v, - half_range, half_range)
    return v


def quantize_linear_scale(scale_value, width, allowed_rate=0.01, step=1.414):
    scale_scale_factor = find_optimal_scale_scale_factor(scale_value, width,
                                                         allowed_rate, step)
    q_scale_value = np.round(scale_value * scale_scale_factor).astype(np.int64)
    return q_scale_value, scale_scale_factor


def find_optimal_scale_scale_factor(scale_value, width, allowed_rate=0.01, step=1.414):
    scale_scale_factor = 1.0

    abs_max = np.max(np.abs(scale_value))
    half_range = 2 ** (width - 1) - 1
    if abs_max > half_range:
        scale_scale_factor = 1.0 * half_range / abs_max

    prev_scale_scale_factor = scale_scale_factor

    while True:
        float_value = np.array(scale_value * scale_scale_factor)
        round_value = np.round(float_value)

        if np.max(np.abs(round_value)) > 2 ** (width - 1) - 1:
            scale_scale_factor = prev_scale_scale_factor
            break

        cont = False
        for i, (f, r) in enumerate(sorted(zip(float_value.reshape([-1]),
                                              round_value.reshape([-1])),
                                          key=lambda x: abs(x[0]), reverse=True)):

            rate = abs(r - f) / (f + 0.0000001)
            if rate > allowed_rate:
                cont = True
                break

            if i >= float_value.size / 10:
                break

        if not cont:
            break

        prev_scale_scale_factor = scale_scale_factor
        scale_scale_factor *= step

    return scale_scale_factor


def quantize_linear_torch_parameter_by_scale_factor(orig_param, scale_factor):
    try:
        import torch
    except:
        raise ImportError('pytorch is required.')

    v = orig_param.data.numpy() * scale_factor
    v = np.round(v).astype(np.int64).astype(np.float32)
    orig_param.data = torch.from_numpy(v)


def to_hist_with_orig_bins(targ_hist, targ_bins, orig_hist, orig_bins):
    targ_v = 0.0
    targ_i = 0
    targ_bin = targ_bins[0]
    ret_hist = np.zeros_like(orig_hist)

    for i, orig_bin in enumerate(orig_bins[:-1]):
        if targ_bin <= orig_bin:
            if targ_i < len(targ_bins) - 1:
                targ_v = targ_hist[targ_i]
                targ_i += 1
                targ_bin = targ_bins[targ_i]
            else:
                targ_v = 0.0
                targ_bin = orig_bin.max() + 1.0

        ret_hist[i] = targ_v

    return ret_hist



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng
import veriloggen

import onnx_matrix_conv2d_linear


act_shape = (1, 4, 4, 3)
weight0_shape = (9, 3, 3, 3)
weight1_shape = (10, 4 * 9)
act_dtype = ng.int8
weight_dtype = ng.int8
stride0 = 1
padding0 = 0
with_batchnorm0 = False
with_batchnorm1 = False
act_func0 = 'ReLU'
act_func1 = 'ReLU'
disable_fusion = False
par_ich = 1
par_och = 1
par_col = 1
par_row = 1
concur_och = None
stationary = 'filter'
chunk_size = 64
axi_datawidth = 32


def test(request, silent=True):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    rslt = onnx_matrix_conv2d_linear.run(act_shape,
                                         weight0_shape, weight1_shape,
                                         act_dtype, weight_dtype,
                                         stride0,
                                         padding0,
                                         with_batchnorm0, with_batchnorm1,
                                         act_func0, act_func1,
                                         disable_fusion,
                                         par_ich, par_och, par_col, par_row,
                                         concur_och, stationary,
                                         chunk_size,
                                         axi_datawidth, silent,
                                         filename=None, simtype=simtype,
                                         outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    verify_rslt = [line for line in rslt.splitlines() if line.startswith('# verify:')][0]
    assert(verify_rslt == '# verify: PASSED')


if __name__ == '__main__':
    rslt = onnx_matrix_conv2d_linear.run(act_shape,
                                         weight0_shape, weight1_shape,
                                         act_dtype, weight_dtype,
                                         stride0,
                                         padding0,
                                         with_batchnorm0, with_batchnorm1,
                                         act_func0, act_func1,
                                         disable_fusion,
                                         par_ich, par_och, par_col, par_row,
                                         concur_och, stationary,
                                         chunk_size,
                                         axi_datawidth, silent=False,
                                         filename='tmp.v',
                                         outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
import numpy as np

from . import util


def normalize(visitor, node):
    input = node.args[0]
    scale = node.args[1]
    bias = node.args[2]
    shamt = node.args[3]

    visitor.visit(input)
    visitor.visit(scale)
    visitor.visit(bias)
    visitor.visit(shamt)

    scale_value = scale.value
    if isinstance(scale_value, (tuple, list)):
        scale_value = np.array(scale_value)

    q_scale_value, scale_scale_factor = util.quantize_linear(scale_value, scale.dtype.width)
    scale.set_value(q_scale_value)
    scale.scale_factor = scale_scale_factor

    bias_value = bias.value
    if isinstance(bias_value, (tuple, list)):
        bias_value = np.array(bias_value)

    q_bias_value = util.quantize_linear_by_scale_factor(
        bias_value, bias.dtype.width, input.scale_factor * scale_scale_factor)
    bias.set_value(q_bias_value)
    bias.scale_factor = input.scale_factor * scale_scale_factor

    init_shamt = max(math.ceil(math.log(np.mean(np.abs(q_scale_value)) * 1.0, 2)), 0)
    q_shamt = find_optimal_shamt_normalize(visitor, node, q_scale_value, q_bias_value,
                                           init_shamt=init_shamt)
    shamt.fill_value = q_shamt
    node.scale_factor = input.scale_factor * scale.scale_factor / (2 ** q_shamt)


def find_optimal_shamt_normalize(visitor, node, scale, bias,
                                 allowed_rate=0.0, range_rate=0.66,
                                 init_shamt=0):

    shamt = init_shamt

    input = node.args[0].eval(visitor.memo, visitor.input_dict)

    if node.dtype.signed:
        _range = round((2 ** (node.dtype.width - 1)) * range_rate)
    else:
        _range = round((2 ** node.dtype.width) * range_rate)

    while True:
        rslt = try_shamt_normalize(node, input, scale, bias, shamt)
        neg_overflow = np.where(rslt <= - _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        pos_overflow = np.where(rslt >= _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        num_overflow = np.sum(neg_overflow + pos_overflow)

        rate = num_overflow / rslt.size
        if rate <= allowed_rate:
            break

        shamt += 1

    visitor.memo[id(node)] = rslt

    return shamt


def try_shamt_normalize(node, x, y, z, shamt):

    import nngen.verify as verify

    name = node.__class__.__name__
    method = getattr(verify, name, None)

    kwargs = {}
    kwargs['dtype'] = node.dtype
    kwargs['sum_dtype'] = node.sum_dtype
    kwargs['name'] = node.name
    kwargs['par'] = node.par

    return method(x, y, z, shamt, **kwargs)


def scaled_add(visitor, node):
    a = node.args[0]
    b = node.args[1]

    visitor.visit(a)
    visitor.visit(b)

    max_scale_factor = max(a.scale_factor, b.scale_factor)
    a_scale_value = max_scale_factor / a.scale_factor
    b_scale_value = max_scale_factor / b.scale_factor

    if max_scale_factor == a.scale_factor:
        q_b_scale_value, b_scale_scale_factor = util.quantize_linear_scale(b_scale_value, 32,
                                                                           allowed_rate=0.001)
        a_scale_value = np.round(b_scale_scale_factor).astype(np.int64)
        q_a_scale_value = a_scale_value
        a_scale_scale_factor = a_scale_value
    elif max_scale_factor == b.scale_factor:
        q_a_scale_value, a_scale_scale_factor = util.quantize_linear_scale(a_scale_value, 32,
                                                                           allowed_rate=0.001)
        b_scale_value = np.round(a_scale_scale_factor).astype(np.int64)
        q_b_scale_value = b_scale_value
        b_scale_scale_factor = b_scale_value

    node.a_scale = int(q_a_scale_value)
    node.b_scale = int(q_b_scale_value)

    init_shamt = max(max(math.ceil(math.log(np.abs(q_a_scale_value) * 1.0, 2)),
                         math.ceil(math.log(np.abs(q_b_scale_value) * 1.0, 2))), 0)
    q_shamt = find_optimal_shamt_scaled_add(visitor, node, q_a_scale_value, q_b_scale_value,
                                            init_shamt=init_shamt)
    node.shamt = q_shamt
    node.scale_factor = max(a.scale_factor * a_scale_scale_factor,
                            b.scale_factor * b_scale_scale_factor) / (2 ** q_shamt)


def find_optimal_shamt_scaled_add(visitor, node, a_scale, b_scale,
                                  allowed_rate=0.0, range_rate=0.66,
                                  init_shamt=0):

    shamt = init_shamt

    a_input = node.args[0].eval(visitor.memo, visitor.input_dict)
    b_input = node.args[1].eval(visitor.memo, visitor.input_dict)

    if node.dtype.signed:
        _range = round((2 ** (node.dtype.width - 1)) * range_rate)
    else:
        _range = round((2 ** node.dtype.width) * range_rate)

    while True:
        rslt = try_shamt_scaled_add(node, a_input, a_scale,
                                    b_input, b_scale, shamt)
        neg_overflow = np.where(rslt <= - _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        pos_overflow = np.where(rslt >= _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        num_overflow = np.sum(neg_overflow + pos_overflow)

        rate = num_overflow / rslt.size
        if rate <= allowed_rate:
            break

        shamt += 1

    visitor.memo[id(node)] = rslt

    return shamt


def try_shamt_scaled_add(node, a, a_scale, b, b_scale, shamt):

    import nngen.verify as verify

    name = node.__class__.__name__
    method = getattr(verify, name, None)

    kwargs = {}
    kwargs['dtype'] = node.dtype
    kwargs['sum_dtype'] = node.sum_dtype
    kwargs['name'] = node.name
    kwargs['par'] = node.par

    return method(a, b, a_scale, b_scale, shamt, **kwargs)


def scaled_concat(visitor, node):
    values = node.args

    for value in values:
        visitor.visit(value)

    max_scale_factor = max(*[value.scale_factor for value in values])
    scale_values = [max_scale_factor / value.scale_factor for value in values]

    new_scale_values = []
    scale_scale_factors = []

    for value, scale_value in zip(values, scale_values):
        if max_scale_factor != value.scale_factor:
            q_scale_value, scale_scale_factor = util.quantize_linear_scale(scale_value, 32,
                                                                           allowed_rate=0.001)
            new_scale_values.append(q_scale_value)
            scale_scale_factors.append(scale_scale_factor)
        else:
            new_scale_values.append(1)
            scale_scale_factors.append(1.0)

    max_scale_scale_factor = max(*scale_scale_factors)

    new_scales = []
    new_scale_scale_factors = []

    for value, new_scale_value, scale_scale_factor in zip(values,
                                                          new_scale_values, scale_scale_factors):

        mag = np.round(max_scale_scale_factor / scale_scale_factor).astype(np.int64)
        scale_value = new_scale_value * mag
        q_scale_value = scale_value
        scale_scale_factor = scale_scale_factor * mag
        new_scales.append(int(scale_value))
        new_scale_scale_factors.append(scale_scale_factor)

    node.scales = new_scales

    init_shamt = max(max(*[math.ceil(math.log(np.abs(new_scale), 2))
                           for new_scale in new_scales]), 0)
    q_shamt = find_optimal_shamt_scaled_concat(visitor, node, new_scales,
                                               init_shamt=init_shamt)
    node.shamt = q_shamt
    node.scale_factor = max(*[value.scale_factor * scale_scale_factor
                              for value, scale_scale_factor in zip(
                                  values, new_scale_scale_factors)]) / (2 ** q_shamt)


def find_optimal_shamt_scaled_concat(visitor, node, scales,
                                     allowed_rate=0.0, range_rate=0.66,
                                     init_shamt=0):

    shamt = init_shamt

    inputs = [arg.eval(visitor.memo, visitor.input_dict) for arg in node.args]

    if node.dtype.signed:
        _range = round((2 ** (node.dtype.width - 1)) * range_rate)
    else:
        _range = round((2 ** node.dtype.width) * range_rate)

    while True:
        rslt = try_shamt_scaled_concat(node, inputs, scales, shamt)
        neg_overflow = np.where(rslt <= - _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        pos_overflow = np.where(rslt >= _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        num_overflow = np.sum(neg_overflow + pos_overflow)

        rate = num_overflow / rslt.size
        if rate <= allowed_rate:
            break

        shamt += 1

    visitor.memo[id(node)] = rslt

    return shamt


def try_shamt_scaled_concat(node, values, scales, shamt):

    import nngen.verify as verify

    name = node.__class__.__name__
    method = getattr(verify, name, None)

    kwargs = {}
    kwargs['axis'] = node.axis
    kwargs['dtype'] = node.dtype
    kwargs['mul_dtype'] = node.mul_dtype
    kwargs['name'] = node.name

    return method(values, scales, shamt, **kwargs)


def scaled_multiply(visitor, node):
    a = node.args[0]
    b = node.args[1]

    visitor.visit(a)
    visitor.visit(b)

    init_shamt = max(a.dtype.width + b.dtype.width - node.dtype.width - 1, 0)
    q_shamt = find_optimal_shamt_scaled_multiply(visitor, node,
                                                 init_shamt=init_shamt)
    node.shamt = q_shamt
    node.scale_factor = a.scale_factor * b.scale_factor / (2 ** q_shamt)


def find_optimal_shamt_scaled_multiply(visitor, node,
                                       allowed_rate=0.0, range_rate=0.33,
                                       init_shamt=0):

    shamt = init_shamt

    a_input = node.args[0].eval(visitor.memo, visitor.input_dict)
    b_input = node.args[1].eval(visitor.memo, visitor.input_dict)

    if node.dtype.signed:
        _range = round((2 ** (node.dtype.width - 1)) * range_rate)
    else:
        _range = round((2 ** node.dtype.width) * range_rate)

    while True:
        rslt = try_shamt_scaled_multiply(node, a_input, b_input, shamt)
        neg_overflow = np.where(rslt <= - _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        pos_overflow = np.where(rslt >= _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        num_overflow = np.sum(neg_overflow + pos_overflow)

        rate = num_overflow / rslt.size
        if rate <= allowed_rate:
            break

        shamt += 1

    visitor.memo[id(node)] = rslt

    return shamt


def try_shamt_scaled_multiply(node, a, b, shamt):

    import nngen.verify as verify

    name = node.__class__.__name__
    method = getattr(verify, name, None)

    kwargs = {}
    kwargs['dtype'] = node.dtype
    kwargs['mul_dtype'] = node.mul_dtype
    kwargs['name'] = node.name
    kwargs['par'] = node.par

    return method(a, b, shamt, **kwargs)


def scaled_div(visitor, node):
    a = node.args[0]
    b = node.args[1]

    visitor.visit(a)
    visitor.visit(b)

    init_shamt = b.dtype.width
    q_shamt = find_optimal_shamt_scaled_div(visitor, node,
                                            init_shamt=init_shamt)
    node.shamt = q_shamt
    node.scale_factor = a.scale_factor / b.scale_factor * (2 ** q_shamt)


def find_optimal_shamt_scaled_div(visitor, node,
                                  allowed_rate=0.0, range_rate=0.95,
                                  init_shamt=0):

    shamt = init_shamt

    a_input = node.args[0].eval(visitor.memo, visitor.input_dict)
    b_input = node.args[1].eval(visitor.memo, visitor.input_dict)
    b_input = np.where(b_input == 0, 1, b_input)

    if node.dtype.signed:
        _range = round((2 ** (node.dtype.width - 1)) * range_rate)
    else:
        _range = round((2 ** node.dtype.width) * range_rate)

    while True:
        rslt = try_shamt_scaled_div(node, a_input, b_input, shamt)
        neg_overflow = np.where(rslt <= - _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        pos_overflow = np.where(rslt >= _range,
                                np.ones_like(rslt), np.zeros_like(rslt))
        num_overflow = np.sum(neg_overflow + pos_overflow)

        rate = num_overflow / rslt.size
        if rate <= allowed_rate:
            break

        shamt -= 1

        if shamt < 0:
            shamt = 0
            break

    visitor.memo[id(node)] = rslt

    return shamt


def try_shamt_scaled_div(node, a, b, shamt):

    import nngen.verify as verify

    name = node.__class__.__name__
    method = getattr(verify, name, None)

    kwargs = {}
    kwargs['dtype'] = node.dtype
    kwargs['div_dtype'] = node.div_dtype
    kwargs['name'] = node.name
    kwargs['par'] = node.par

    return method(a, b, shamt, **kwargs)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections

import nngen.storage as storage
import nngen.operator as operator
import nngen.dtype_list as dtype_list

from . import util
from . import flatten
from . import reshape


def Gemm(visitor, node,
         batchnorm_scale=None, batchnorm_bias=None, act_func=None):

    # input, filter
    srcs = []

    for i, src in enumerate(node.input):
        src_node = util.search_node_from_model(visitor.model, src)

        if (i == 0 and src_node.op_type == 'Flatten' and
                len(visitor.consumers[src]) == 1):

            src_obj = flatten.Flatten(visitor, src_node, no_transpose=True)
            srcs.append(src_obj)
            continue

        if (i == 0 and src_node.op_type == 'Reshape' and
                len(visitor.consumers[src]) == 1):

            shape = visitor.visit(src_node.input[1])
            if len(shape) == 2:
                src_obj = reshape.Reshape(visitor, src_node, no_transpose=True)
                srcs.append(src_obj)
                continue

        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    input = srcs[0]
    filter = srcs[1]

    orig_shape = input.get_original_shape()
    orig_layout = input.get_original_layout()
    orig_onnx_layout = input.get_original_onnx_layout()

    if orig_layout is None:
        pass
    elif orig_layout == orig_onnx_layout:
        pass
    else:
        # The weight layout of Gemm is identical to nngen.matmul.
        # However, Gemm assumes values before the Reshape operator have the different layout.
        # (Almost ONNX models usually have 'NCHW' layouts).
        # Thus the weight layout is transposed.

        shape = ([filter.shape[0]] +
                 [orig_shape[orig_layout.index(s)] for s in orig_onnx_layout[1:]])
        reshape_value = filter.value.reshape(shape)
        perm = [orig_onnx_layout.index(s) for s in orig_layout]
        transpose_value = reshape_value.transpose(perm)
        new_value = transpose_value.reshape([filter.shape[0], -1])
        filter.value = new_value

    bias = srcs[2] if len(srcs) > 2 else None

    name = util.get_name(node)

    scale_name = '_'.join(['onnx', name, 'gemm.scale'])
    scale_dtype = visitor.default_scale_dtype
    scale_shape = batchnorm_scale.shape if batchnorm_scale is not None else (1,)
    scale = storage.variable(dtype=scale_dtype, shape=scale_shape, name=scale_name)
    scale_value = batchnorm_scale if batchnorm_scale is not None else [1]
    scale.set_value(scale_value)
    visitor.variables[scale_name] = scale

    if bias is None and batchnorm_bias is not None:
        bias_name = '_'.join(['onnx', name, 'gemm.bias'])
        bias_dtype = visitor.default_bias_dtype
        bias_shape = batchnorm_bias.shape
        bias = storage.variable(dtype=bias_dtype, shape=bias_shape, name=bias_name)
        bias_value = batchnorm_bias / batchnorm_scale
        bias.set_value(bias_value)
        visitor.variables[bias_name] = bias

    elif bias is not None and batchnorm_bias is not None:
        bias.dtype = visitor.default_bias_dtype
        bias_value = batchnorm_bias / batchnorm_scale + bias.value
        bias.set_value(bias_value)

    elif bias is not None:
        bias.dtype = visitor.default_bias_dtype

    # rshift_out_name = '_'.join(['onnx, name, 'gemm.rshift_out'])
    #rshift_out_width = filter.dtype.width
    #rshift_out_dtype = dtype_list.dtype_int(rshift_out_width, signed=False)
    #rshift_out_shape = (1,)
    #rshift_out = storage.variable(dtype=scale_dtype, shape=scale_shape, name=rshift_out_name)
    #visitor.variables[rshift_out_name] = rshift_out
    rshift_out = 0

    if name in visitor.value_dtypes:
        dtype = visitor.value_dtypes[name]
    else:
        dtype = visitor.default_operator_dtype

    if dtype.width >= 16:
        sum_dtype = dtype_list.dtype_int(dtype.width * 4)
    else:
        sum_dtype = dtype_list.int32

    args = [input, filter]

    kwargs = collections.OrderedDict()
    kwargs['bias'] = bias
    kwargs['scale'] = scale
    kwargs['transposed_a'] = False
    kwargs['transposed_b'] = True
    kwargs['rshift_out'] = rshift_out
    kwargs['act_func'] = act_func
    kwargs['dtype'] = dtype
    kwargs['sum_dtype'] = sum_dtype
    kwargs['name'] = name

    c = operator.matmul(*args, **kwargs)

    return c



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections
import numpy as np

import nngen.operator as operator

from . import util


def Upsample(visitor, node):

    mode = 'nearest'
    scale_value = 1.0

    for attribute in node.attribute:
        if attribute.name == 'mode':
            mode = attribute.s.decode()

        # deprecated attribute since Upsample-9
        if attribute.name == 'scale':
            scale_value = attribute.f

    if mode != 'nearest':
        raise ValueError("Upsampling mode must be 'nearest', not '%s'." % mode)

    if round(scale_value) != scale_value:
        raise ValueError("Upsampling factor must be a multiple of integer, not %f." %
                         scale_value)

    scale_value = round(scale_value)

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    srcs = [util.optimize_to_raw_value(src) for src in srcs]

    input = srcs[0]

    if len(input.shape) != 4:
        raise ValueError("not supported shape: %s" % str(tuple(shape)))

    # transpose data layout to nngen-compatible format
    input = util.transpose_layout(input, visitor.nngen_input_layout, visitor.onnx_input_layout)

    name = util.get_name(node)

    if len(srcs) > 1:
        factors = srcs[1]
        if not isinstance(factors, (np.ndarray, np.float, np.int, float, int)):
            raise TypeError("Upsampling factor must be constant, not %s." % str(type(factors)))

        if not isinstance(factors, np.ndarray):
            factors = np.array(factors)

        if np.not_equal(np.round(factors), factors).any():
            raise ValueError("Upsampling factor must be a multiple of integer, not %s." %
                             str(factors))

        if len(factors) == 4:
            factors = [int(round(factors[visitor.onnx_input_layout.index(l)]))
                       for l in visitor.nngen_input_layout]
        else:
            factors = np.array([int(round(factors.reshape([-1])[0]))] * 4)

    else:
        factors = np.array([scale_value] * 4)

    if name in visitor.value_dtypes:
        dtype = visitor.value_dtypes[name]
    else:
        dtype = visitor.default_operator_dtype

    kwargs = collections.OrderedDict()
    kwargs['factors'] = factors
    kwargs['dtype'] = dtype
    kwargs['name'] = name

    c = operator.upsampling2d(input, **kwargs)
    c.layout = visitor.nngen_input_layout
    c.onnx_layout = visitor.onnx_input_layout

    return c



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np


def concat(values, axis, dtype=None, name=None):
    return np.concatenate(values, axis=axis)
